#!/usr/bin/env python
# coding: utf-8
"""Flask-based server sample code"""
import argparse
from flask import Flask
from flask_compress import Compress  # type: ignore
from flask_cors import CORS  # type: ignore
import os

# Disable libhdf5 file locking since h5grove is only reading files
# This needs to be done before any import of h5py, so before h5grove import
os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"

from h5grove.flaskutils import BLUEPRINT as h5grove_blueprint  # noqa


def parser_fn():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "-p", "--port", type=int, default=8888, help="Port the server is listening on"
    )
    parser.add_argument(
        "--ip", default="localhost", help="IP the server is listening on"
    )
    parser.add_argument(
        "--compression", action="store_true", help="Enable HTTP compression"
    )
    parser.add_argument(
        "--basedir",
        default=".",
        help="Base directory from which to retrieve HDF5 files",
    )
    return parser


if __name__ == "__main__":
    parser = parser_fn()
    options = parser.parse_args()
    # Create Flask application
    app = Flask(__name__)

    # Enable cross-origin resource sharing, see https://flask-cors.readthedocs.io
    CORS(app)

    # Enable compression, see https://github.com/colour-science/flask-compress
    if options.compression:
        Compress(app)

    # Configure h5grove default endpoints
    app.config["H5_BASE_DIR"] = os.path.abspath(options.basedir)
    app.register_blueprint(h5grove_blueprint)

    # Start server
    app.run(port=options.port, host=options.ip)



from typing import Dict, Generic, Optional, Sequence, TypeVar, Union
import h5py
import numpy as np

try:
    import hdf5plugin  # noqa: F401
except ImportError:
    pass

from .models import LinkResolution, Selection
from .utils import (
    attr_metadata,
    get_array_stats,
    get_filters,
    get_entity_from_file,
    hdf_path_join,
    get_dataset_slice,
    sorted_dict,
)


class EntityContent:
    """Base content for an entity."""

    type = "other"

    def __init__(self, path: str):
        self._path = path

    def metadata(self) -> Dict[str, str]:
        """Entity metadata

        :returns: {"name": str, "type": str}
        """
        return {"name": self.name, "type": self.type}

    @property
    def name(self) -> str:
        """Entity name. Last member of the path."""
        return self._path.split("/")[-1]

    @property
    def path(self) -> str:
        """Path in the file."""
        return self._path


class ExternalLinkContent(EntityContent):
    type = "external_link"

    def __init__(self, path: str, link: h5py.ExternalLink):
        super().__init__(path)
        self._target_file = link.filename
        self._target_path = link.path

    def metadata(self, depth=None):
        """External link metadata

        :returns: {"name": str, "target_file": str, "target_path": str, "type": str}
        """
        return sorted_dict(
            ("target_file", self._target_file),
            ("target_path", self._target_path),
            *super().metadata().items(),
        )

    @property
    def target_file(self) -> str:
        """The target file of the link"""
        return self._target_file

    @property
    def target_path(self) -> str:
        """The target path of the link (in the target file)"""
        return self._target_path


class SoftLinkContent(EntityContent):
    type = "soft_link"

    def __init__(self, path: str, link: h5py.SoftLink) -> None:
        super().__init__(path)
        self._target_path = link.path
        """ The target path of the link """

    def metadata(self, depth=None):
        """
        :returns: {"name": str, "target_path": str, "type": str}
        """
        return sorted_dict(
            ("target_path", self._target_path), *super().metadata().items()
        )

    @property
    def target_path(self) -> str:
        """The target path of the link"""
        return self._target_path


T = TypeVar("T", h5py.Dataset, h5py.Datatype, h5py.Group)


class ResolvedEntityContent(EntityContent, Generic[T]):
    """Content for a link that can be resolved into a h5py entity"""

    def __init__(self, path: str, h5py_entity: T):
        super().__init__(path)
        self._h5py_entity = h5py_entity
        """ Resolved h5py entity """

    def attributes(self, attr_keys: Optional[Sequence[str]] = None):
        """Attributes of the h5py entity. Can be filtered by keys."""
        if attr_keys is None:
            return dict((*self._h5py_entity.attrs.items(),))

        return dict((key, self._h5py_entity.attrs[key]) for key in attr_keys)

    def metadata(self, depth=None):
        """
        :returns: {"attributes": AttributeMetadata, "name": str, "type": str}
        """
        attribute_names = sorted(self._h5py_entity.attrs.keys())
        return sorted_dict(
            (
                "attributes",
                [
                    attr_metadata(self._h5py_entity.attrs.get_id(k))
                    for k in attribute_names
                ],
            ),
            *super().metadata().items(),
        )


class DatasetContent(ResolvedEntityContent[h5py.Dataset]):
    type = "dataset"

    def metadata(self, depth=None):
        """
        :returns: {"attributes": AttributeMetadata, chunks": tuple, "dtype": str, "filters": tuple, "shape": tuple, "name": str, "type": str}
        """
        return sorted_dict(
            ("chunks", self._h5py_entity.chunks),
            ("dtype", self._h5py_entity.dtype.str),
            ("filters", get_filters(self._h5py_entity)),
            ("shape", self._h5py_entity.shape),
            *super().metadata().items(),
        )

    def data(self, selection: Selection = None, flatten: bool = False):
        """Dataset data. Supports slicing through selection."""
        result = get_dataset_slice(self._h5py_entity, selection)

        # Do not flatten scalars nor h5py.Empty
        if flatten and isinstance(result, np.ndarray):
            return np.ravel(result)

        return result

    def data_stats(
        self, selection: Selection = None
    ) -> Dict[str, Union[float, int, None]]:
        """Statistics on the data. Providing a selection will compute stats only on the selected slice.

        :param selection: NumPy-like indexing to define a selection as a slice
        :returns: {"strict_positive_min": number | None, "positive_min": number | None, "min": number | None, "max": number | None, "mean": number | None, "std": number | None}
        """
        data = self._get_finite_data(selection)

        return get_array_stats(data)

    def _get_finite_data(self, selection: Selection) -> np.ndarray:
        data = np.array(self.data(selection), copy=False)  # So it works with scalars

        if not np.issubdtype(data.dtype, np.floating):
            return data

        mask = np.isfinite(data)
        if np.all(mask):
            return data

        return data[mask]


class GroupContent(ResolvedEntityContent[h5py.Group]):
    type = "group"

    def __init__(self, path: str, h5py_entity: h5py.Group, h5file: h5py.File):
        super().__init__(path, h5py_entity)
        self._h5file = h5file
        """ File in which the entity was resolved. This is needed to resolve child entity. """

    def _get_child_metadata_content(self, depth=0):
        return [
            create_content(
                self._h5file, hdf_path_join(self._path, child_path)
            ).metadata(depth)
            for child_path in self._h5py_entity.keys()
        ]

    def metadata(self, depth: int = 1):
        """Metadata of the group. Recursively includes child metadata if depth > 0.

        :parameter depth: The level of child metadata resolution.
        :returns: {"attributes": AttributeMetadata, "children": ChildMetadata, "name": str, "type": str}
        """
        if depth <= 0:
            return super().metadata()

        return sorted_dict(
            ("children", self._get_child_metadata_content(depth - 1)),
            *super().metadata().items(),
        )


def create_content(
    h5file: h5py.File,
    path: Optional[str],
    resolve_links: LinkResolution = LinkResolution.ONLY_VALID,
):
    """
    Factory function to get entity content from a HDF5 file.
    This handles external/soft link resolution and dataset decompression.

    :param h5file: An open HDF5 file containing the entity
    :param path: Path to the entity in the file.
    :param resolve_links: Tells which external and soft links should be resolved. Defaults to resolving only valid links.
    :raises h5grove.utils.PathError: If the path cannot be found in the file
    :raises h5grove.utils.LinkError: If a link cannot be resolved when resolve_links is set to LinkResolution.ALL.
    :raises TypeError: If encountering an unsupported h5py entity
    """
    if path is None:
        path = "/"

    entity = get_entity_from_file(h5file, path, resolve_links)

    if isinstance(entity, h5py.ExternalLink):
        return ExternalLinkContent(path, entity)

    if isinstance(entity, h5py.SoftLink):
        return SoftLinkContent(path, entity)

    if isinstance(entity, h5py.Dataset):
        return DatasetContent(path, entity)

    if isinstance(entity, h5py.Group):
        return GroupContent(path, entity, h5file)

    if isinstance(entity, h5py.Datatype):
        return ResolvedEntityContent(path, entity)

    raise TypeError(f"h5py type {type(entity)} not supported")



import io
from numbers import Number
from typing import Any, Callable, Dict, Optional, Sequence, Union
import numpy as np
import orjson
import h5py
import tifffile

from .utils import sanitize_array


def bin_encode(array: Sequence[Number]) -> bytes:
    """Sanitize an array and convert it to bytes.

    :param array: Data to convert
    """
    sanitized_array = sanitize_array(array)

    return sanitized_array.tobytes()


def orjson_default(o: Any) -> Union[list, float, str, None]:
    """Converts Python objects to JSON-serializable objects.

    :raises TypeError: if the object is not supported."""
    if isinstance(o, np.number) and o.dtype.kind == "f" and o.itemsize > 8:
        # Force conversion of float >64bits to native float even if it means losing precision
        return float(o)
    if isinstance(o, (np.generic, np.ndarray)):
        return o.tolist()
    if isinstance(o, complex):
        return [o.real, o.imag]
    if isinstance(o, h5py.Empty):
        return None
    if isinstance(o, bytes):
        return o.decode()
    if isinstance(o, (h5py.RegionReference, h5py.Reference)):
        return str(o)
    raise TypeError


def orjson_encode(content: Any, default: Optional[Callable] = None) -> bytes:
    """Encode in JSON using orjson.

    :param: content: Content to encode
    :param default: orjson default (https://github.com/ijl/orjson#default)
    """
    if default is None:
        default = orjson_default

    return orjson.dumps(content, default=default, option=orjson.OPT_SERIALIZE_NUMPY)


def csv_encode(data: np.ndarray) -> bytes:
    """Encodes a NumPy array in CSV.

    :param: data: NumPy array to encode
    """
    with io.BytesIO() as buffer:
        np.savetxt(buffer, data, delimiter=",")
        return buffer.getvalue()


def npy_encode(data: np.ndarray) -> bytes:
    """Encodes a NumPy array in NPY.

    :param: data: NumPy array to encode
    """
    with io.BytesIO() as buffer:
        np.save(buffer, data)
        return buffer.getvalue()


def tiff_encode(data: np.ndarray) -> bytes:
    """Encodes a NumPy array in TIFF. The data should be 2D.

    :param: data: NumPy array to encode
    """
    with io.BytesIO() as buffer:
        tifffile.imwrite(buffer, data, photometric="minisblack")
        return buffer.getvalue()


class Response:
    content: bytes
    """ Encoded `content` as bytes """
    headers: Dict[str, str]
    """ Associated headers """

    def __init__(self, content: bytes, headers: Dict[str, str]):
        self.content = content
        self.headers = {**headers, "Content-Length": str(len(content))}


def encode(content, encoding: Optional[str] = "json") -> Response:
    """Encode content in given encoding.

    Warning: Not all encodings supports all types of content.

    :param content: Content to encode
    :param encoding:
        - `json` (default)
        - `bin`: nD array/scalars in bytes
        - `csv`: nD arrays in downloadable csv files
        - `npy`: nD arrays in downloadable npy files
        - `tiff`: 2D arrays in downloadable TIFF files
    :returns: A Response object containing content and headers
    :raises ValueError: If encoding is not among the ones above.
    """
    if encoding in ("json", None):
        return Response(
            orjson_encode(content),
            headers={"Content-Type": "application/json"},
        )

    if encoding == "bin":
        return Response(
            bin_encode(content),
            headers={
                "Content-Type": "application/octet-stream",
            },
        )

    if encoding == "csv":
        return Response(
            csv_encode(content),
            headers={
                "Content-Type": "text/csv",
                "Content-Disposition": 'attachment; filename="data.csv"',
            },
        )

    if encoding == "npy":
        return Response(
            npy_encode(content),
            headers={
                "Content-Type": "application/octet-stream",
                "Content-Disposition": 'attachment; filename="data.npy"',
            },
        )

    if encoding == "tiff":
        return Response(
            tiff_encode(content),
            headers={
                "Content-Type": "image/tiff",
                "Content-Disposition": 'attachment; filename="data.tiff"',
            },
        )

    raise ValueError(f"Unsupported encoding {encoding}")



from enum import Enum
from typing import Tuple, Union
import h5py

H5pyEntity = Union[
    h5py.Dataset, h5py.Datatype, h5py.ExternalLink, h5py.Group, h5py.SoftLink
]

Selection = Union[str, int, slice, Tuple[Union[slice, int], ...], None]


class LinkResolution(str, Enum):
    NONE = "none"  # Links are not resolved whatever their status
    ONLY_VALID = (
        "only_valid"  # Links are only resolved if valid and unresolved if broken
    )
    ALL = "all"  # Links are resolved no matter their status. The resolution of broken links raises LinkError



import h5py
from numbers import Number
from os.path import basename
import numpy as np
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from .models import H5pyEntity, LinkResolution, Selection


class NotFoundError(Exception):
    pass


class PathError(NotFoundError):
    pass


class LinkError(NotFoundError):
    pass


def attr_metadata(attrId: h5py.h5a.AttrID) -> dict:
    return {"dtype": attrId.dtype.str, "name": attrId.name, "shape": attrId.shape}


def get_entity_from_file(
    h5file: h5py.File,
    path: str,
    resolve_links: LinkResolution = LinkResolution.ONLY_VALID,
) -> H5pyEntity:
    if path == "/":
        return h5file[path]

    link = h5file.get(path, getlink=True)

    if link is None:
        raise PathError(f"{path} is not a valid path in {basename(h5file.filename)}")

    if isinstance(link, h5py.ExternalLink) or isinstance(link, h5py.SoftLink):
        if resolve_links == LinkResolution.NONE:
            return link

        try:
            return h5file[path]
        except (OSError, KeyError):
            if resolve_links == LinkResolution.ONLY_VALID:
                return link

            raise LinkError(
                f"Cannot resolve {link} at {path} of {basename(h5file.filename)}"
            )

    return h5file[path]


def parse_slice(dataset: h5py.Dataset, slice_str: str) -> Tuple[Union[slice, int], ...]:
    if "," not in slice_str:
        return (parse_slice_member(slice_str, dataset.shape[0]),)

    slice_members = slice_str.split(",")

    if len(slice_members) > dataset.ndim:
        raise TypeError(
            f"{slice_str} is a {len(slice_members)}d slice while the dataset is {dataset.ndim}d"
        )

    return tuple(
        parse_slice_member(s, dataset.shape[i]) for i, s in enumerate(slice_members)
    )


def parse_slice_member(slice_member: str, max_dim: int) -> Union[slice, int]:
    if ":" not in slice_member:
        return int(slice_member)

    slice_params = slice_member.split(":")
    if len(slice_params) == 2:
        start, stop = slice_params

        return slice(
            int(start) if start != "" else 0, int(stop) if stop != "" else max_dim
        )

    if len(slice_params) == 3:
        start, stop, step = slice_params

        return slice(
            int(start) if start != "" else 0,
            int(stop) if stop != "" else max_dim,
            int(step) if step != "" else 1,
        )

    raise TypeError(f"{slice_member} is not a valid slice")


def sorted_dict(*args: Tuple[str, Any]):
    return dict(sorted(args))


def _sanitize_dtype(dtype: np.dtype) -> np.dtype:
    """Convert dtype to a dtype supported by js-numpy-parser.

    See https://github.com/ludwigschubert/js-numpy-parser

    :raises ValueError: For unsupported array dtype
    """
    if dtype.kind not in ("f", "i", "u"):
        raise ValueError("Unsupported array type")

    # Convert to little endian
    result = dtype.newbyteorder("<")

    if result.kind == "i" and result.itemsize > 4:
        return np.dtype("<i4")  # int64 -> int32

    if result.kind == "u" and result.itemsize > 4:
        return np.dtype("<u4")  # uint64 -> uint32

    if result.kind == "f" and result.itemsize < 4:
        return np.dtype("<f4")

    if result.kind == "f" and result.itemsize > 8:
        return np.dtype("<f8")

    return result


def sanitize_array(array: Sequence[Number], copy: bool = True) -> np.ndarray:
    """Ensure array save as .npy can be read back by js-numpy-parser.

    See https://github.com/ludwigschubert/js-numpy-parser

    :param array: Array to sanitize
    :param copy: Set to False to avoid copy if possible
    :raises ValueError: For unsupported array dtype
    """
    ndarray = np.array(array, copy=False)
    return np.array(ndarray, copy=copy, order="C", dtype=_sanitize_dtype(ndarray.dtype))


def get_array_stats(data: np.ndarray) -> Dict[str, Union[float, int, None]]:
    if data.size == 0:
        return {
            "strict_positive_min": None,
            "positive_min": None,
            "min": None,
            "max": None,
            "mean": None,
            "std": None,
        }

    cast = float if np.issubdtype(data.dtype, np.floating) else int
    strict_positive_data = data[data > 0]
    positive_data = data[data >= 0]
    return {
        "strict_positive_min": cast(np.min(strict_positive_data))
        if strict_positive_data.size != 0
        else None,
        "positive_min": cast(np.min(positive_data))
        if positive_data.size != 0
        else None,
        "min": cast(np.min(data)),
        "max": cast(np.max(data)),
        "mean": cast(np.mean(data)),
        "std": cast(np.std(data)),
    }


def hdf_path_join(prefix, suffix):
    if prefix == "/":
        return f"/{suffix}"

    return f'{prefix.rstrip("/")}/{suffix}'


def parse_bool_arg(query_arg: Union[str, None], fallback: bool) -> bool:
    if query_arg is None:
        return fallback

    return query_arg.lower() != "false"


def parse_link_resolution_arg(
    raw_query_arg: Union[str, None], fallback: LinkResolution
) -> LinkResolution:
    if raw_query_arg is None:
        return fallback

    query_arg = raw_query_arg.lower()

    # Checking for "true"/"false" to keep the same behaviour as when the arg was a boolean
    if query_arg in ("true", LinkResolution.ALL):
        return LinkResolution.ALL

    if query_arg in ("false", LinkResolution.NONE):
        return LinkResolution.NONE

    if query_arg == LinkResolution.ONLY_VALID:
        return LinkResolution.ONLY_VALID

    raise ValueError(
        f"{raw_query_arg} is not a valid value for link resolution. Accepted values are: {LinkResolution.ALL}, f{LinkResolution.NONE} or {LinkResolution.ONLY_VALID}"
    )


def get_dataset_slice(dataset: h5py.Dataset, selection: Selection):
    if selection is None:
        return dataset[()]

    if isinstance(selection, str):
        return dataset[parse_slice(dataset, selection)]

    return dataset[selection]


def get_filters(
    dataset: h5py.Dataset,
) -> Optional[List[Dict[str, Union[int, str]]]]:
    property_list = dataset.id.get_create_plist()

    n_filters = property_list.get_nfilters()
    if n_filters <= 0:
        return None

    return [get_filter_info(property_list.get_filter(i)) for i in range(n_filters)]


def get_filter_info(
    filter: Tuple[int, int, Tuple[int, ...], str]
) -> Dict[str, Union[int, str]]:
    # https://api.h5py.org/h5p.html#h5py.h5p.PropDCID.get_filter
    (filter_id, _, _, name) = filter

    return {"id": filter_id, "name": name}



"""Helpers for usage with `Flask <https://flask.palletsprojects.com/>`_"""
from flask import abort, Blueprint, current_app, request, Response, Request
import h5py
import os
from typing import Any, Callable, Mapping, Optional


from .content import create_content, DatasetContent, ResolvedEntityContent
from .encoders import encode
from .models import LinkResolution
from .utils import NotFoundError, parse_bool_arg, parse_link_resolution_arg


__all__ = [
    "attr_route",
    "data_route",
    "meta_route",
    "stats_route",
    "URL_RULES",
    "BLUEPRINT",
]


def make_encoded_response(content, format_arg: Optional[str]) -> Response:
    """Prepare flask Response according to format"""
    h5grove_response = encode(content, format_arg)
    response = Response(h5grove_response.content)
    response.headers.update(h5grove_response.headers)
    return response


def get_content(
    h5file: h5py.File,
    path: Optional[str],
    resolve_links: LinkResolution = LinkResolution.ONLY_VALID,
):
    """Gets contents if path is in file. Raises 404 otherwise"""
    try:
        return create_content(h5file, path, resolve_links)
    except NotFoundError as e:
        abort(404, str(e))


def get_filename(a_request: Request) -> str:
    file_path = a_request.args.get("file")
    if file_path is None:
        raise KeyError("File argument is required")

    full_file_path = os.path.join(current_app.config["H5_BASE_DIR"], file_path)
    if not os.path.isfile(full_file_path):
        abort(404, "File not found!")

    return full_file_path


def attr_route():
    """`/attr/` endpoint handler"""
    filename = get_filename(request)
    path = request.args.get("path")
    format_arg = request.args.get("format")
    attr_keys = (
        request.args.getlist("attr_keys") if "attr_keys" in request.args else None
    )

    with h5py.File(filename, mode="r") as h5file:
        content = get_content(h5file, path)
        assert isinstance(content, ResolvedEntityContent)
        return make_encoded_response(content.attributes(attr_keys), format_arg)


def data_route():
    """`/data/` endpoint handler"""
    filename = get_filename(request)
    path = request.args.get("path")
    selection = request.args.get("selection")
    format_arg = request.args.get("format")
    flatten = parse_bool_arg(request.args.get("flatten"), fallback=False)

    with h5py.File(filename, mode="r") as h5file:
        content = get_content(h5file, path)
        assert isinstance(content, DatasetContent)
        return make_encoded_response(content.data(selection, flatten), format_arg)


def meta_route():
    """`/meta/` endpoint handler"""
    filename = get_filename(request)
    path = request.args.get("path")
    format_arg = request.args.get("format")
    resolve_links = parse_link_resolution_arg(
        request.args.get("resolve_links", None),
        fallback=LinkResolution.ONLY_VALID,
    )

    with h5py.File(filename, mode="r") as h5file:
        content = get_content(h5file, path, resolve_links)
        return make_encoded_response(content.metadata(), format_arg)


def stats_route():
    """`/stats/` endpoint handler"""
    filename = get_filename(request)
    path = request.args.get("path")
    selection = request.args.get("selection")
    format_arg = request.args.get("format")

    with h5py.File(filename, mode="r") as h5file:
        content = get_content(h5file, path)
        assert isinstance(content, DatasetContent)
        return make_encoded_response(content.data_stats(selection), format_arg)


URL_RULES = {
    "/attr/": attr_route,
    "/data/": data_route,
    "/meta/": meta_route,
    "/stats/": stats_route,
}
"""Mapping of Flask URL endpoints to handlers"""


BLUEPRINT = Blueprint("h5grove", __name__)
"""Blueprint of h5grove endpoints.

It relies on `H5_BASE_DIR` being defined in the app config.
"""


def _init_blueprint(blueprint: Blueprint, url_rules: Mapping[str, Callable[[], Any]]):
    for url, view_func in url_rules.items():
        blueprint.add_url_rule(url, view_func=view_func)


_init_blueprint(BLUEPRINT, URL_RULES)



#!/usr/bin/env python
# coding: utf-8
"""Tornado-based server sample code"""
import argparse
import os
import tornado.web
import tornado.ioloop

# Disable libhdf5 file locking since h5grove is only reading files
# This needs to be done before any import of h5py, so before h5grove import
os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"

from h5grove.tornadoutils import get_handlers  # noqa


def parser_fn():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "-p", "--port", type=int, default=8888, help="Port the server is listening on"
    )
    parser.add_argument(
        "--ip", default="localhost", help="IP the server is listening on"
    )
    parser.add_argument(
        "--basedir",
        default=".",
        help="Base directory from which to retrieve HDF5 files",
    )
    return parser


if __name__ == "__main__":
    parser = parser_fn()
    options = parser.parse_args()

    base_dir = os.path.abspath(options.basedir)

    app = tornado.web.Application(get_handlers(base_dir, allow_origin="*"), debug=True)
    app.listen(options.port, options.ip)
    print(f"App is listening on {options.ip}:{options.port} serving from {base_dir}...")
    tornado.ioloop.IOLoop.current().start()



"""Helpers for usage with `Tornado <https://www.tornadoweb.org>`_"""
import os
from typing import Any, Optional
import h5py
from tornado.web import RequestHandler, MissingArgumentError, HTTPError

from .content import DatasetContent, ResolvedEntityContent, create_content
from .encoders import encode
from .models import LinkResolution
from .utils import NotFoundError, parse_bool_arg, parse_link_resolution_arg


__all__ = [
    "BaseHandler",
    "AttributeHandler",
    "DataHandler",
    "MetadataHandler",
    "StatisticsHandler",
    "get_handlers",
]


class BaseHandler(RequestHandler):
    """Base class for h5grove handlers"""

    def initialize(self, base_dir: str, allow_origin: Optional[str] = None) -> None:
        self.base_dir = base_dir
        self.allow_origin = allow_origin

    def get(self):
        file_path = self.get_query_argument("file")
        if file_path is None:
            raise MissingArgumentError("file")

        path = self.get_query_argument("path", None, strip=False)
        format_arg = self.get_query_argument("format", None)

        full_file_path = os.path.join(self.base_dir, file_path)
        if not os.path.isfile(full_file_path):
            raise HTTPError(status_code=404, reason="File not found!")

        with h5py.File(full_file_path, "r") as h5file:
            try:
                content = self.get_content(h5file, path)
            except NotFoundError as e:
                raise HTTPError(status_code=404, reason=str(e))

        response = encode(content, format_arg)

        for key, value in response.headers.items():
            self.set_header(key, value)

        self.write(response.content)
        self.finish()

    def get_content(self, h5file: h5py.File, path: Optional[str]):
        raise NotImplementedError

    def prepare(self):
        if self.allow_origin is not None:
            self.set_header("Access-Control-Allow-Origin", self.allow_origin)

    def write_error(self, status_code: int, **kwargs: Any) -> None:
        self.prepare()
        return super().write_error(status_code, **kwargs)


class AttributeHandler(BaseHandler):
    """/attr/ endpoint handler"""

    def get_content(self, h5file: h5py.File, path: Optional[str]):
        content = create_content(h5file, path)
        assert isinstance(content, ResolvedEntityContent)

        attr_keys = self.get_query_arguments("attr_keys", strip=False)
        # get_query_arguments returns an empty list if `attr_keys` is not present
        return content.attributes(attr_keys if len(attr_keys) > 0 else None)


class DataHandler(BaseHandler):
    """/data/ endpoint handler"""

    def get_content(self, h5file: h5py.File, path: Optional[str]):
        selection = self.get_query_argument("selection", None)
        flatten = parse_bool_arg(
            self.get_query_argument("flatten", None), fallback=False
        )

        content = create_content(h5file, path)
        assert isinstance(content, DatasetContent)
        return content.data(selection, flatten)


class MetadataHandler(BaseHandler):
    """/meta/ endpoint handler"""

    def get_content(self, h5file: h5py.File, path: Optional[str]):
        resolve_links = parse_link_resolution_arg(
            self.get_query_argument("resolve_links", None),
            fallback=LinkResolution.ONLY_VALID,
        )
        content = create_content(h5file, path, resolve_links)
        return content.metadata()


class StatisticsHandler(BaseHandler):
    """/stats/ endpoint handler"""

    def get_content(self, h5file: h5py.File, path: Optional[str]):
        selection = self.get_query_argument("selection", None)

        content = create_content(h5file, path)
        assert isinstance(content, DatasetContent)
        return content.data_stats(selection)


# TODO: Setting the return type raises mypy errors
def get_handlers(base_dir: Optional[str], allow_origin: Optional[str] = None):
    """Build h5grove handlers (`/attr`, `/data`, `/meta` and `/stats`).

    :param base_dir: Base directory from which the HDF5 files will be served
    :param allow_origin: Allowed origins for CORS
    :return type: List[Tuple[str, BaseHandler, dict]]
    """
    init_args = {"base_dir": base_dir, "allow_origin": allow_origin}
    return [
        (r"/attr/.*", AttributeHandler, init_args),
        (r"/data/.*", DataHandler, init_args),
        (r"/meta/.*", MetadataHandler, init_args),
        (r"/stats/.*", StatisticsHandler, init_args),
    ]
